# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command Line Interface (CLI) for ASReview project."""
import argparse
import inspect
import sys
from importlib.metadata import metadata
from itertools import groupby

from asreview import __version__
from asreview.utils import _entry_points


def _execute_entry_point(entry, args):
    if inspect.isclass(entry):
        entry().execute(args)
    else:
        entry(args)


def main():
    # Get the available entry points.
    base_entries = _entry_points(group="asreview.entry_points")
    base_entries_internal = _entry_points(group="asreview.entry_points_internal")

    if (
        len(sys.argv) > 1
        and not sys.argv[1].startswith("-")
        and not (
            sys.argv[1] in base_entries.names
            or sys.argv[1] in base_entries_internal.names
        )
    ):
        raise ValueError(f"'{sys.argv[1]}' is not a valid subcommand.")

    elif len(sys.argv) > 1 and sys.argv[1] in base_entries.names:
        entry = base_entries[sys.argv[1]].load()
        _execute_entry_point(entry, sys.argv[2:])
    elif len(sys.argv) > 1 and sys.argv[1] in base_entries_internal.names:
        entry = base_entries_internal[sys.argv[1]].load()
        _execute_entry_point(entry, sys.argv[2:])

    else:
        description_subcommands = ""

        for name, dist_entry_points in groupby(
            base_entries,
            lambda e: e.dist.name,
        ):
            description = metadata(name)["Summary"]
            version = metadata(name)["Version"]
            description_subcommands += f"\n[{name} {version}] - {description}\n"

            for entry in dist_entry_points:
                description_subcommands += f"\t{entry.name}\n"

        parser = argparse.ArgumentParser(
            prog="asreview",
            formatter_class=argparse.RawTextHelpFormatter,
            description=metadata("asreview")["Summary"],
        )
        parser.add_argument(
            "subcommand",
            nargs="?",
            default=None,
            help=f"The subcommand to launch. Available commands:\n\n"
            f"{description_subcommands}",
        )

        parser.add_argument(
            "-V",
            "--version",
            action="version",
            version="%(prog)s {version}".format(version=__version__),
        )

        args, _ = parser.parse_known_args()

        parser.print_help()


if __name__ == "__main__":
    main()



__all__ = ["AuthTool"]

import argparse
import json
import os
import sys
from argparse import RawTextHelpFormatter
from uuid import UUID

from sqlalchemy import create_engine
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker

from asreview.project import ASReviewProject
from asreview.utils import asreview_path
from asreview.webapp.authentication.models import Project
from asreview.webapp.authentication.models import User
from asreview.webapp.authentication.models import create_database_and_tables

DEFAULT_DATABASE_URI = f"sqlite:///{str(asreview_path())}/asreview.production.sqlite"

DB_URI_HELP = (
    "URI of the database. By default, the value is given by the environment "
    "variable ASREVIEW_LAB_SQLALCHEMY_DATABASE_URI. If not set, the default "
    "is 'asreview.production.sqlite' in the ASReview folder."
)


def auth_parser():
    parser = argparse.ArgumentParser(
        prog="auth-tool",
        description="""
Tool to create and fill a database to authenticate the ASReview application.
The tool can be used to convert existing project data to be used in an
authenticated setup.
        """,
        formatter_class=RawTextHelpFormatter,
        epilog="Use -h or --help on all subcommands to view the available options.",
    )

    sub_parser = parser.add_subparsers(help="The following options are available:")

    # CREATE DB
    create_db_par = sub_parser.add_parser(
        "create-db",
        help="Create the database necessary to authenticate the ASReview app.",
    )

    create_db_par.add_argument(
        "-d",
        "--db-uri",
        type=str,
        default=None,
        help=DB_URI_HELP,
    )

    # ADD USERS
    user_par = sub_parser.add_parser("add-users", help="Add users into the database.")

    user_par.add_argument(
        "-d",
        "--db-uri",
        type=str,
        default=None,
        help=DB_URI_HELP,
    )

    user_par.add_argument(
        "-j",
        "--json",
        type=str,
        help="JSON string that contains a list with user account data.",
    )

    # RESET PASSWORD USER
    user_reset_password_par = sub_parser.add_parser(
        "reset-password", help="Reset password of a single user account."
    )

    user_reset_password_par.add_argument(
        "-d",
        "--db-uri",
        type=str,
        default=None,
        help=DB_URI_HELP,
    )

    # LIST USERS
    list_users_par = sub_parser.add_parser(
        "list-users",
        help="List user accounts.",
    )

    list_users_par.add_argument(
        "-d",
        "--db-uri",
        type=str,
        default=None,
        help=DB_URI_HELP,
    )

    # LIST PROJECTS
    list_projects_par = sub_parser.add_parser(
        "list-projects",
        help="List project info from all projects in the ASReview folder.",
    )

    list_projects_par.add_argument(
        "-j",
        "--json",
        action="store_true",
        help="Create JSON string to connect existing projects with users.",
    )

    # LINK PROJECTS TO USERS
    link_par = sub_parser.add_parser(
        "link-projects", help="Link projects to user accounts."
    )

    link_par.add_argument(
        "-j",
        "--json",
        type=str,
        help="Use a JSON string to link projects to users.",
    )

    link_par.add_argument(
        "-d",
        "--db-uri",
        type=str,
        default=None,
        help=DB_URI_HELP,
    )

    return parser


def verify_id(id):
    try:
        UUID(id)
        return True
    except ValueError:
        return False


def insert_user(session, entry):
    """Inserts a dictionary containing user data
    into the database."""
    # create a user object
    user = User(
        entry["email"].lower(),
        email=entry["email"].lower(),
        name=entry["name"],
        affiliation=entry["affiliation"],
        password=entry["password"],
        confirmed=True,
    )
    try:
        session.add(user)
        session.commit()
        print(f"User with email {user.email} created.")
        return True
    except IntegrityError:
        session.rollback()
        sys.stderr.write(f"User with identifier {user.email} already exists")
        return False


def insert_project(session, project):
    # get owner and project id
    owner_id = project["owner_id"]
    project_id = project["project_id"]

    # check if this project was already in the database under
    # the old project id
    db_project = (
        session.query(Project).filter(Project.project_id == project_id).one_or_none()
    )
    if db_project is None:
        # create new record
        session.add(Project(owner_id=owner_id, project_id=project_id))
    else:
        # update record (project_id must be the same)
        db_project.owner_id = owner_id
    # commit
    session.commit()
    print("Project data is stored.")
    return True


def get_users(session):
    return session.query(User).all()


def get_user_by_id(session, user_id):
    return session.query(User).filter_by(id=user_id).one()


class AuthTool:
    def execute(self, argv):
        parser = auth_parser()
        args = parser.parse_args(argv)

        self.args = args
        self.argv = argv

        # create a conn object for the database if needed (database
        # is not needed when the only command is "list-projects", all
        # other commands need a database session)
        if self.argv != ["list-projects"]:
            self.uri = (
                getattr(self.args, "db_uri", False)
                or os.environ.get("SQLALCHEMY_DATABASE_URI", False)
                or os.environ.get("ASREVIEW_LAB_SQLALCHEMY_DATABASE_URI", False)
                or DEFAULT_DATABASE_URI
            )
            Session = sessionmaker()
            engine = create_engine(self.uri)
            Session.configure(bind=engine)
            self.session = Session()

        if "create-db" in argv:
            self.create_database()
        elif "add-users" in argv:
            self.add_users()
        elif "reset-password" in argv:
            self.reset_password()
        elif "list-users" in argv:
            self.list_users()
        elif "list-projects" in argv:
            self.list_projects()
        elif "link-projects" in argv:
            self.link_projects()

    def create_database(self):
        engine = create_engine(self.uri)
        create_database_and_tables(engine)
        print(f"...Database created, URI: {self.uri}")

    def add_users(self):
        if self.args.json is not None:
            entries = json.loads(self.args.json)
            # try to insert entries into the database
            for entry in entries:
                insert_user(self.session, entry)
        else:
            self.enter_users()

    def reset_password(self):
        """Resets password for a user."""
        # get user accounts
        user_ids = [u.id for u in get_users(self.session)]
        print("\nReset password of:")
        self.list_users()
        # get user id
        id = self._ensure_valid_value_for(
            "Id of user account",
            lambda x: x.isnumeric() and int(x) in user_ids,
            hint="Id number must be a number and exist.",
        )
        # get new password, since this is not hidden I don't see
        # the point of confirmation
        password = self._ensure_valid_value_for(
            "New password",
            User.valid_password,
            "Use 8 or more characters with a mix of letters, numbers & symbols.",  # noqa
        )
        # id and password are guaranteed, get user and reset password
        user = get_user_by_id(self.session, int(id))
        user.reset_password(password)
        self.session.commit()
        print(f"Password of {user.name} has been reset.")

    def _ensure_valid_value_for(self, name, validation_function, hint=""):
        """Prompt user for validated input."""
        while True:
            value = input(f"{name}: ")
            if validation_function(value):
                return value
            else:
                sys.stderr.write(hint + "\n")

    def enter_users(self):
        while True:
            new_user = input("Enter a new user [Y/n]? ")
            if new_user == "Y":
                email = self._ensure_valid_value_for(
                    "Email address (required)",
                    User.valid_email,
                    "Entered email address is not recognized as a valid email address.",  # noqa
                )
                name = self._ensure_valid_value_for(
                    "Full name (required)",
                    lambda x: bool(x) and len(x) > 2,
                    "Full name must contain more than 2 characters.",
                )
                affiliation = input("Affiliation: ")
                password = self._ensure_valid_value_for(
                    "Password (required)",
                    User.valid_password,
                    "Use 8 or more characters with a mix of letters, numbers & symbols.",  # noqa
                )

                insert_user(
                    self.session,
                    {
                        "email": email,
                        "name": name,
                        "affiliation": affiliation,
                        "password": password,
                    },
                )
            else:
                break

        return True

    def _print_project(self, project):
        print(f"\n* {project['folder']}")
        print(f"\tversion: {project['version']}")
        print(f"\tid: {project['project_id']}")
        print(f"\tname: {project['name']}")
        print(f"\tauthors: {project['authors']}")
        print(f"\tcreated: {project['created']}")

    def _print_user(self, user):
        if bool(user.affiliation):
            postfix = f", {user.affiliation}"
        else:
            postfix = ""
        print(f" {user.id} - {user.email} ({user.name}){postfix}")

    def _get_projects(self):
        projects = [f for f in asreview_path().glob("*") if f.is_dir()]
        result = []
        for folder in projects:
            project = ASReviewProject(folder)

            # Raise a RuntimeError if the project version is too low.
            if project.config.get("version").startswith("0."):
                id = project.config.get("id")
                message = f"""Version of project with id {id} is too old,
                please upgrade first before using this tool."""
                raise RuntimeError(message)

            result.append(
                {
                    "folder": folder.name,
                    "version": project.config.get("version"),
                    "project_id": project.config.get("id"),
                    "name": project.config.get("name"),
                    "authors": project.config.get("authors"),
                    "created": project.config.get("datetimeCreated"),
                    "owner_id": 0,
                }
            )
        return result

    def list_users(self):
        users = get_users(self.session)
        print()
        for user in users:
            self._print_user(user)
        print()

    def list_projects(self):
        projects = self._get_projects()
        if self.args.json:
            # dump the data twice to create a string
            # that can be loaded again by the tool.
            print(json.dumps(json.dumps(projects)))
        else:
            [self._print_project(p) for p in projects]
            if len(projects) > 0:
                print()

    def _generate_project_links(self):
        result = []
        # get users and projects
        users = get_users(self.session)
        user_ids = [u.id for u in users]
        projects = self._get_projects()
        # print projects
        for project in projects:
            self._print_project(project)
            print("Who's the owner of this project?")
            print("--------------------------------")
            for user in users:
                self._print_user(user)
            id = None
            # and ask who the owner is
            while True:
                id = input("Enter the ID number of the owner: ")
                try:
                    if isinstance(id, str):
                        id = id.replace(".", "")
                    id = int(id)
                    if id not in user_ids:
                        print("Entered ID does not exists, try again.")
                    else:
                        insert_project(
                            self.session,
                            {"project_id": project["project_id"], "owner_id": id},
                        )
                        break
                except ValueError:
                    sys.stderr.write("Entered ID is not a number, please try again.")
        return result

    def link_projects(self):
        # bulk JSON vs interactive
        if self.args.json is not None:
            projects = json.loads(self.args.json)
            # enter data in the database
            for project in projects:
                insert_project(self.session, project)
        else:
            self._generate_project_links()



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "BaseDataGroup",
    "BaseDataSet",
    "BenchmarkDataGroup",
    "DatasetManager",
    "DatasetNotFoundError",
    "NaturePublicationDataGroup",
    "SynergyDataGroup",
    "SynergyDataSet",
]

import json
import socket
import tempfile
import warnings
from abc import ABC
from abc import abstractmethod
from pathlib import Path
from urllib.error import URLError
from urllib.request import urlopen
from urllib.request import urlretrieve

import synergy_dataset as sd

from asreview.io import CSVReader
from asreview.utils import _entry_points
from asreview.utils import _get_filename_from_url
from asreview.utils import is_iterable


class DatasetNotFoundError(Exception):
    pass


def _download_from_metadata(url):
    """Download metadata to dataset."""

    try:
        with urlopen(url, timeout=10) as f:
            meta_data = json.loads(f.read().decode())
    except URLError as e:
        if isinstance(e.reason, socket.timeout):
            raise Exception("Connection time out.")
        raise e

    datasets = []
    for data in meta_data.values():
        # raise error on versioned datasets
        if "type" in data and data["type"] == "versioned":
            raise ValueError("Datasets of type 'versioned' are deprecated")

        datasets.append(BaseDataSet(**data))

    return datasets


class BaseDataSet:
    def __init__(
        self,
        dataset_id,
        filepath=None,
        title=None,
        description=None,
        authors=None,
        topic=None,
        link=None,
        reference=None,
        img_url=None,
        license=None,
        year=None,
        aliases=[],
        **kwargs,
    ):
        """Base class for metadata of dataset.

        A BaseDataSet is a class with metadata about a (labeled)
        dataset used in ASReview LAB. The dataset can be used via
        the frontend or via command line interface.

        In general, a BaseDataSet is part of a group (BaseDataGroup).

        Examples
        --------

        The following example simulates a dataset with dataset_id
        'cord19'. The name of the group is 'covid'.

        >>> asreview simulate covid:cord_19

        Parameters
        ----------
        dataset_id: str
            Identifier of the dataset. The value is a alphanumeric
            string used to indentify the dataset via the command line
            interface. Example: 'groupname:DATASET_ID' where DATASET_ID
            is the value of dataset_id.
        filepath: str
            Path to file or URL to the dataset. See
            asreview.readthedocs.io/{URL} for information about valid
            datasets.
        title: str
            Title of the dataset.
        description: str
            Description of the dataset. Optional.
        authors: list
            Authors of the dataset. Optional.
        topic: str
            Topics of the dataset. Optional.
        link: str
            Link to a website or additional information.
        reference: str
            (Academic) reference describing the dataset. Optional.
        license: str
            License of the dataset. Optional
        year: str
            Year of publication of the dataset. Optional.
        img_url: str
            Image for display in graphical interfaces. Optional.
        aliases: list
            Additional identifiers for the dataset_id. This can be
            useful for long of complex dataset_id's. Optional.

        """

        self.dataset_id = dataset_id
        self.filepath = filepath
        self.title = title
        self.description = description
        self.authors = authors
        self.topic = topic
        self.link = link
        self.reference = reference
        self.license = license
        self.year = year
        self.img_url = img_url
        self.aliases = aliases
        self.kwargs = kwargs

    def __str__(self):
        return f"<BaseDataSet dataset_id='{self.dataset_id}' title='{self.title}'>"

    def __dict__(self):
        return {
            "dataset_id": self.dataset_id,
            "filepath": self.filepath,
            "title": self.title,
            "description": self.description,
            "authors": self.authors,
            "topic": self.topic,
            "link": self.link,
            "reference": self.reference,
            "license": self.license,
            "year": self.year,
            "img_url": self.img_url,
            "aliases": self.aliases,
            **self.kwargs,
        }

    @property
    def reader(self):
        return None

    @property
    def filename(self):
        if not hasattr(self, "_filename"):
            self._filename = _get_filename_from_url(self.filepath)

        return self._filename

    def to_file(self, path):
        # todo return without store
        urlretrieve(self.filepath, path)


class BaseDataGroup(ABC):
    def __init__(self, *datasets):
        """Group of datasets.

        Group containing one or more datasets.

        Parameters
        ----------
        *datasets:
            One or more datasets.
        """
        self.datasets = list(datasets)

    @property
    @abstractmethod
    def group_id(cls):
        pass

    @property
    @abstractmethod
    def description(cls):
        pass

    def __str__(self):
        return f"<BaseDataGroup group_id='{self.group_id}'>"

    def __dict__(self):
        return {d.dataset_id: d for d in self.datasets}

    def append(self, dataset):
        """Append dataset to group.

        dataset: asreview.datasets.BaseDataSet
            A asreview BaseDataSet-like object.
        """
        if not issubclass(dataset, BaseDataSet):
            raise ValueError("Expected BaseDataSet or subclass of BaseDataSet.")
        self.datasets.append(dataset)

    def find(self, dataset_id):
        """Find dataset in the group.

        Parameters
        ----------
        dataset_id: str
            Identifier of the dataset to look for. It can also be one
            of the aliases. Case insensitive.

        Returns
        -------
        asreview.datasets.BaseDataSet:
            Returns base dataset with the given dataset_id.
        """
        results = []
        for d in self.datasets:
            if dataset_id.lower() == d.dataset_id.lower() or dataset_id.lower() in [
                a.lower() for a in d.aliases
            ]:
                results.append(d)

        if len(results) > 1:
            raise ValueError(
                f"Broken dataset group '{self.group_id}' containing multiple"
                f" datasets with the same name/alias '{dataset_id}'."
            )
        elif len(results) == 1:
            return results[0]

        raise DatasetNotFoundError(f"Dataset {dataset_id} not found")


class DatasetManager:
    @property
    def groups(self):
        return list(_entry_points(group="asreview.datasets").names)

    def find(self, dataset_id):
        """Find a dataset.

        Parameters
        ----------
        dataset_id: str, iterable
            Look for this term in aliases within any dataset. A group can
            be specified by setting dataset_id to 'group_id:dataset_id'.
            This can be helpful if the dataset_id is not unique.
            The dataset_id can also be a non-string iterable, in which case
            a list will be returned with all terms.
            Dataset_ids should not contain semicolons (:).
            Return None if the dataset could not be found.

        Returns
        -------
        BaseDataSet:
            Return the dataset with dataset_id.
        """
        # If dataset_id is a non-string iterable, return a list.
        if is_iterable(dataset_id):
            return [self.find(x) for x in dataset_id]

        # If dataset_id is a valid path, create a dataset from it.
        if Path(dataset_id).is_file():
            return BaseDataSet(dataset_id)

        dataset_id = str(dataset_id)

        # get installed dataset groups
        dataset_groups = _entry_points(group="asreview.datasets")

        # Split into group/dataset if possible.
        split_dataset_id = dataset_id.split(":")
        if len(split_dataset_id) == 2:
            data_group = split_dataset_id[0]
            split_dataset_id = split_dataset_id[1]
            if data_group in self.groups:
                return dataset_groups[data_group].load()().find(split_dataset_id)

        # Look through all available/installed groups for the name.
        all_results = {}
        for data_group in dataset_groups:
            try:
                all_results[data_group.name] = data_group.load()().find(dataset_id)
            except Exception:
                # don't raise error on loading entry point
                pass

        # If we have multiple results, throw an error.
        if len(all_results) > 1:
            raise ValueError(
                f"Multiple datasets found: {list(all_results)}."
                "Use DATAGROUP:DATASET format to specify which one"
                " you want."
            )

        if len(all_results) == 1:
            return list(all_results.values())[0]

        # Could not find dataset
        raise DatasetNotFoundError(f"Dataset {dataset_id} not found")

    def list(self, include=None, exclude=None, serialize=True, raise_on_error=False):
        """List the available datasets.

        Parameters
        ----------
        include: str, iterable
            List of groups to include
        exclude: str, iterable
            List of groups to exclude from all groups.
        serialize: bool
            Make returned list serializable.
        raise_on_error: bool
            Raise error when entry point can't be loaded.

        Returns
        -------
        list:
            List with datasets as values.
        """

        if include is not None and exclude is not None:
            raise ValueError("Cannot exclude groups when include is not None.")

        if include is not None:
            if not is_iterable(include):
                include = [include]
            groups = include
        elif exclude is not None:
            exclude = exclude if is_iterable(exclude) else [exclude]
            groups = list(set(self.groups) - set(exclude))
        else:
            groups = self.groups.copy()

        dataset_groups = _entry_points(group="asreview.datasets")

        group_list = []
        for group in groups:
            try:
                group_list.append(dataset_groups[group].load()())
            except Exception as err:
                # don't raise error on loading entry point
                if raise_on_error:
                    raise err

        if serialize:
            dataset_list_ser = []
            for data_group in group_list:
                try:
                    group_ser = []
                    for dataset in data_group.datasets:
                        group_ser.append(dataset.__dict__())
                    dataset_list_ser.append(
                        {
                            "group_id": data_group.group_id,
                            "description": data_group.description,
                            "datasets": group_ser,
                        }
                    )
                except Exception as err:
                    # don't raise error on loading entry point
                    if raise_on_error:
                        raise err

            return dataset_list_ser

        return group_list


class NaturePublicationDataGroup(BaseDataGroup):
    """Datasets used in the paper Van de Schoot et al. 2020."""

    group_id = "benchmark-nature"
    description = (
        "Datasets used in the validation paper published"
        " in Nature Machine Intelligence (van de Schoot et al. 2021)"
    )

    def __init__(self):
        meta_file = "https://raw.githubusercontent.com/asreview/paper-asreview/master/index_v1.json"  # noqa
        datasets = _download_from_metadata(meta_file)

        super(NaturePublicationDataGroup, self).__init__(*datasets)


class SynergyDataSet(BaseDataSet):
    @property
    def filename(self):
        return self.dataset_id + ".csv"

    @property
    def reader(self):
        return CSVReader

    def to_file(self, path=None):
        # download, build, and store to local file
        try:
            return sd.Dataset(self.dataset_id).to_frame().to_csv(path)
        except FileNotFoundError:
            tmp_synergy_folder = tempfile.mkdtemp()
            sd.download_raw_subset(self.dataset_id, path=tmp_synergy_folder)

            for d in sd.iter_datasets(path=tmp_synergy_folder):
                if d.name == self.dataset_id:
                    return d.to_frame().to_csv(path)

        raise ValueError("Synergy dataset not found")


class SynergyDataGroup(BaseDataGroup):
    """Datasets available in the SYNERGY dataset."""

    group_id = "synergy"
    description = "SYNERGY datasets (asreview.ai/synergy)"

    def __init__(self):
        # The following code was used to generate the metadata
        #
        # import synergy_dataset as sd
        # from pprint import pprint
        # meta_synergy = {}
        # for x in sd.iter_datasets():
        #     meta_synergy[x.name] = {
        #         "title": x.metadata["publication"]["display_name"],
        #         "authors": x.cite.split(",")[0] + " et al.",
        #         "topic": x.metadata
        #            ["data"]["concepts"]["included"][0]["display_name"],
        #         "link": "https://doi.org/10.34894/HE6NAQ",
        #         "reference": x.metadata["publication"]["doi"],
        #         "license": "See Synergy dataset",
        #         "year": x.metadata["publication"]["publication_year"]
        #     }
        # pprint(meta_synergy)

        synergy_metadata = {
            "Appenzeller-Herzog_2019": {
                "authors": "Appenzeller‐Herzog et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1111/liv.14179",
                "title": "Comparative effectiveness of common "
                "therapies for Wilson disease: A "
                "systematic review and meta‐analysis of "
                "controlled studies",
                "topic": "Medicine",
                "year": 2019,
            },
            "Bos_2018": {
                "authors": "Bos et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1016/j.jalz.2018.04.007",
                "title": "Cerebral small vessel disease and the risk of "
                "dementia: A systematic review and meta‐analysis of "
                "population‐based evidence",
                "topic": "Medicine",
                "year": 2018,
            },
            "Brouwer_2019": {
                "authors": "Brouwer et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1016/j.cpr.2019.101773",
                "title": "Psychological theories of depressive relapse and "
                "recurrence: A systematic review and meta-analysis "
                "of prospective studies",
                "topic": "Psychology",
                "year": 2019,
            },
            "Chou_2003": {
                "authors": "Chou et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1016/j.jpainsymman.2003.03.003",
                "title": "Comparative efficacy and safety of long-acting oral "
                "opioids for chronic non-cancer pain: a systematic "
                "review",
                "topic": "Medicine",
                "year": 2003,
            },
            "Chou_2004": {
                "authors": "Chou et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1016/j.jpainsymman.2004.05.002",
                "title": "Comparative efficacy and safety of skeletal muscle "
                "relaxants for spasticity and musculoskeletal "
                "conditions: a systematic review",
                "topic": "Medicine",
                "year": 2004,
            },
            "Donners_2021": {
                "authors": "Donners et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1007/s40262-021-01042-w",
                "title": "Pharmacokinetics and Associated Efficacy of "
                "Emicizumab in Humans: A Systematic Review",
                "topic": "Medicine",
                "year": 2021,
            },
            "Hall_2012": {
                "authors": "Hall et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1109/tse.2011.103",
                "title": "A Systematic Literature Review on Fault Prediction "
                "Performance in Software Engineering",
                "topic": "Computer science",
                "year": 2012,
            },
            "Jeyaraman_2020": {
                "authors": "Jeyaraman et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1177/1947603520951623",
                "title": "Does the Source of Mesenchymal Stem Cell Have an "
                "Effect in the Management of Osteoarthritis of "
                "the Knee? Meta-Analysis of Randomized Controlled "
                "Trials",
                "topic": "Medicine",
                "year": 2020,
            },
            "Leenaars_2019": {
                "authors": "Leenaars et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.5334/jcr.183",
                "title": "Sleep and Microdialysis: An Experiment and a "
                "Systematic Review of Histamine and Several Amino "
                "Acids",
                "topic": "Psychology",
                "year": 2019,
            },
            "Leenaars_2020": {
                "authors": "Leenaars et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.3390/ani10061047",
                "title": "A Systematic Review Comparing Experimental Design "
                "of Animal and Human Methotrexate Efficacy Studies "
                "for Rheumatoid Arthritis: Lessons for the "
                "Translational Value of Animal Studies",
                "topic": "Medicine",
                "year": 2020,
            },
            "Meijboom_2021": {
                "authors": "Meijboom et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1007/s40259-021-00508-4",
                "title": "Patients Retransitioning from Biosimilar TNFα "
                "Inhibitor to the Corresponding Originator After "
                "Initial Transitioning to the Biosimilar: A "
                "Systematic Review",
                "topic": "Medicine",
                "year": 2021,
            },
            "Menon_2022": {
                "authors": "Menon et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1080/10408444.2022.2082917",
                "title": "The methodological rigour of systematic reviews in "
                "environmental health",
                "topic": "Medicine",
                "year": 2022,
            },
            "Moran_2021": {
                "authors": "Moran et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1111/brv.12655",
                "title": "Poor nutritional condition promotes high‐risk "
                "behaviours: a systematic review and meta‐analysis",
                "topic": "Biology",
                "year": 2021,
            },
            "Muthu_2021": {
                "authors": "Muthu et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1097/brs.0000000000003645",
                "title": "Fragility Analysis of Statistically Significant "
                "Outcomes of Randomized Control Trials in Spine "
                "Surgery",
                "topic": "Medicine",
                "year": 2021,
            },
            "Nelson_2002": {
                "authors": "Nelson et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1001/jama.288.7.872",
                "title": "Postmenopausal Hormone Replacement Therapy",
                "topic": "Medicine",
                "year": 2002,
            },
            "Oud_2018": {
                "authors": "Oud et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1177/0004867418791257",
                "title": "Specialized psychotherapies for adults with borderline "
                "personality disorder: A systematic review and "
                "meta-analysis",
                "topic": "Psychology",
                "year": 2018,
            },
            "Radjenovic_2013": {
                "authors": "Radjenović et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1016/j.infsof.2013.02.009",
                "title": "Software fault prediction metrics: A systematic "
                "literature review",
                "topic": "Computer science",
                "year": 2013,
            },
            "Sep_2021": {
                "authors": "Sep et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1371/journal.pone.0249102",
                "title": "The rodent object-in-context task: A systematic review "
                "and meta-analysis of important variables",
                "topic": "Psychology",
                "year": 2021,
            },
            "Smid_2020": {
                "authors": "Smid et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1080/10705511.2019.1577140",
                "title": "Bayesian Versus Frequentist Estimation for Structural "
                "Equation Models in Small Sample Contexts: A "
                "Systematic Review",
                "topic": "Computer science",
                "year": 2020,
            },
            "Walker_2018": {
                "authors": "Walker et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1016/j.envint.2017.12.032",
                "title": "Human and animal evidence of potential "
                "transgenerational inheritance of health effects: An "
                "evidence map and state-of-the-science evaluation",
                "topic": "Biology",
                "year": 2018,
            },
            "Wassenaar_2017": {
                "authors": "Wassenaar et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1289/ehp1233",
                "title": "Systematic Review and Meta-Analysis of "
                "Early-Life Exposure to Bisphenol A and "
                "Obesity-Related Outcomes in Rodents",
                "topic": "Medicine",
                "year": 2017,
            },
            "Wolters_2018": {
                "authors": "Wolters et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1016/j.jalz.2018.01.007",
                "title": "Coronary heart disease, heart failure, and the "
                "risk of dementia: A systematic review and "
                "meta‐analysis",
                "topic": "Medicine",
                "year": 2018,
            },
            "van_Dis_2020": {
                "authors": "van Dis et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1001/jamapsychiatry.2019.3986",
                "title": "Long-term Outcomes of Cognitive Behavioral Therapy "
                "for Anxiety-Related Disorders",
                "topic": "Psychology",
                "year": 2020,
            },
            "van_de_Schoot_2018": {
                "authors": "van de Schoot et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1080/00273171.2017.1412293",
                "title": "Bayesian PTSD-Trajectory Analysis with "
                "Informed Priors Based on a Systematic "
                "Literature Search and Expert Elicitation",
                "topic": "Psychology",
                "year": 2018,
            },
            "van_der_Valk_2021": {
                "authors": "Valk et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1111/obr.13376",
                "title": "Cross‐sectional relation of long‐term "
                "glucocorticoids in hair with anthropometric "
                "measurements and their possible determinants: "
                "A systematic review and meta‐analysis",
                "topic": "Medicine",
                "year": 2021,
            },
            "van_der_Waal_2022": {
                "authors": "van der Waal et al.",
                "license": "See Synergy dataset",
                "link": "https://doi.org/10.34894/HE6NAQ",
                "reference": "https://doi.org/10.1016/j.jgo.2022.09.012",
                "title": "A meta-analysis on the role older adults with "
                "cancer favour in treatment decision making",
                "topic": "Medicine",
                "year": 2022,
            },
        }

        datasets = [SynergyDataSet(k, **v) for k, v in synergy_metadata.items()]

        super(SynergyDataGroup, self).__init__(*datasets)


class BenchmarkDataGroup(BaseDataGroup):
    """Datasets available in the benchmark platform.
    Deprecated
    """

    group_id = "benchmark"
    description = "DEPRECATED: Datasets available in the online benchmark platform"

    def __init__(self):

        warnings.warn(
            "The use of 'benchmark' datasets is deprecated, "
            "use SYNERGY dataset instead. For more information, see "
            "https://github.com/asreview/synergy-dataset.",
            category=UserWarning
        )

        meta_file = "https://raw.githubusercontent.com/asreview/systematic-review-datasets/master/index_v1.json"  # noqa
        datasets = _download_from_metadata(meta_file)

        super(BenchmarkDataGroup, self).__init__(*datasets)



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from asreview import ASReviewData


@pytest.mark.xfail(raises=ValueError, reason="Bad record_id")
def test_bad_record_id():
    data_fp = Path("tests", "demo_data", "generic_bad_record_id.csv")
    as_data = ASReviewData.from_file(data_fp)
    assert len(np.unique(as_data.df.index.values)) == len(as_data)


def test_record_id():
    data_fp = Path("tests", "demo_data", "record_id.csv")
    as_data = ASReviewData.from_file(data_fp)

    # test is labels are numpy array
    assert isinstance(as_data.labels, np.ndarray)

    # test is index name is record_id
    assert as_data.df.index.name == "record_id"


def test_column_names_with_spaces():
    data_fp = Path("tests", "demo_data", "generic.csv")
    as_data = ASReviewData.from_file(data_fp)

    data_fp_bad_cols = Path(
        "tests", "demo_data", "generic_column_names_with_spaces.csv"
    )
    as_data_bad_cols = ASReviewData.from_file(data_fp_bad_cols)

    assert_frame_equal(
        as_data.df[["title", "abstract"]], as_data_bad_cols.df[["title", "abstract"]]
    )


def test_asdata_init():
    data_fp = Path("tests", "demo_data", "generic.csv")

    # data via pandas
    df = pd.read_csv(data_fp)
    df.index.name = "record_id"
    as_data_init = ASReviewData(df)

    # data via classmethod
    as_data = ASReviewData.from_file(data_fp)

    assert_frame_equal(as_data_init.df, as_data.df)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps

from flask import current_app
from flask import jsonify
from flask_login import current_user

from asreview.project import ASReviewProject
from asreview.project import ProjectNotFoundError
from asreview.project import get_project_path
from asreview.project import get_projects
from asreview.project import is_project
from asreview.webapp.authentication.models import Project


def project_authorization(f):
    """Decorator function that checks if current user can access
    a project in an authenticated situation"""

    @wraps(f)
    def decorated_function(project_id, *args, **kwargs):
        if current_app.config.get("LOGIN_DISABLED", False):
            project_path = get_project_path(project_id)
            if not is_project(project_path):
                raise ProjectNotFoundError(f"Project '{project_id}' not found")
            project = ASReviewProject(project_path, project_id=project_id)
            return f(project, *args, **kwargs)

        # find the project
        project = Project.query.filter(Project.project_id == project_id).one_or_none()

        # raise ProjectNotFoundError if not exists
        if project is None:
            raise ProjectNotFoundError(f"Project '{project_id}' not found")

        # if there is a project, check if permissiton
        all_users = set([project.owner] + project.collaborators)
        if current_user not in all_users:
            return jsonify({"message": "no permission"}), 403

        project_path = get_project_path(project_id)
        if not is_project(project_path):
            raise ProjectNotFoundError(f"Project '{project_id}' not found")
        project = ASReviewProject(project_path, project_id=project_id)

        return f(project, *args, **kwargs)

    return decorated_function


def current_user_projects(f):
    """Decorator that returns all authenticated projects"""

    @wraps(f)
    def decorated_function(*args, **kwargs):
        if current_app.config.get("LOGIN_DISABLED", False):
            projects = get_projects(None)
        else:
            # authenticated with User accounts
            user_db_projects = list(current_user.projects) + list(
                current_user.involved_in
            )
            projects = get_projects(
                [project.project_path for project in user_db_projects]
            )

        return f(projects, *args, **kwargs)

    return decorated_function



from asreview.utils import get_random_state


def test_automatic_seed():
    random_state = get_random_state()
    seed = random_state.seed
    assert isinstance(seed, int)
    new_random_state = get_random_state(seed)
    assert new_random_state.seed == seed


def test_integer_seed():
    random_state = get_random_state(42)
    assert random_state.seed == 42


def test_state_seed():
    random_state = get_random_state(42)
    assert random_state.seed == 42
    new_random_state = get_random_state(random_state)
    assert new_random_state.seed == 42



from pathlib import Path

import rispy
from pytest import mark

from asreview import ASReviewData
from asreview.utils import is_url


@mark.parametrize(
    "test_file,n_lines,ignore_col",
    [
        ("_baseline.ris", 100, []),
        ("embase.csv", 6, ["keywords"]),
        ("embase_newpage.csv", 6, ["keywords"]),
        ("embase.ris", 6, []),
        ("generic.csv", 2, []),
        ("generic_semicolon.csv", 2, []),
        ("generic_tab.csv", 2, []),
        ("generic_tab.tab", 2, []),
        ("generic_tab.tsv", 2, []),
        ("generic_labels.csv", 6, []),
        ("pubmed_zotero.ris", 6, []),
        ("pubmed_endnote.txt", 6, []),
        ("scopus.ris", 6, []),
        ("ovid_zotero.ris", 6, []),
        ("proquest.ris", 6, []),
        ("https://osf.io/download/fg93a/", 38, []),
    ],
)
def test_reader(test_file, n_lines, ignore_col):
    if is_url(test_file):
        fp = test_file
    else:
        fp = Path("tests", "demo_data", test_file)

    as_data = ASReviewData.from_file(fp)
    assert len(as_data) == n_lines

    cols = ["title", "abstract", "authors", "keywords"]
    cols = [col for col in cols if col not in ignore_col]
    # if labels is not None:
    #     cols.append('included')
    #     assert np.array_equal(as_data.labels, labels)

    for col in cols:
        values = as_data.get(col)
        assert len(values) == n_lines


@mark.parametrize(
    "record_i,included",
    [
        # Single line record
        (0, 1),
        (1, 0),
        (2, -1),
        (3, -1),
        # Single line record with additional notes, label first
        (4, 1),
        (5, 0),
        (6, -1),
        # Single line record with additional notes, label in the middle
        (7, 1),
        (8, 0),
        (9, -1),
        # Single line record with additional notes, label last
        (10, 1),
        (11, 0),
        (12, -1),
        # Multiline record, label first
        (13, 1),
        (14, 0),
        (15, -1),
        # Multiline record, label in the middle
        (16, 1),
        (17, 0),
        (18, -1),
        # Multiline record, label last
        (19, 1),
        (20, 0),
        (21, -1),
        # Multiline record, with additional notes, label first
        (22, 1),
        (23, 0),
        (24, -1),
        # Multiline record, with additional notes, label in the middle
        (25, 1),
        (26, 0),
        (27, -1),
        # Multiline record, with additional notes, label last
        (28, 1),
        (29, 0),
        (30, -1),
        # No notes tag present
        (31, -1),
    ],
)
def test_asreview_labels_ris(record_i, included):
    fp = Path("tests", "demo_data", "baseline_tag-notes_labels.ris")
    as_data = ASReviewData.from_file(fp)
    assert as_data.record(record_i, by_index=True).included == included


def test_multiline_tags_ris():
    fp = Path("tests", "demo_data", "baseline_tag_and_field_definitions_lists.ris")
    entries = rispy.load(fp, encoding="utf-8")
    assert entries[0]["notes"] == ["Notes 1", "Notes 2"]


def test_nan_values_ris():
    fp = Path("tests", "demo_data", "baseline_empty_values.ris")
    as_data = ASReviewData.from_file(fp)

    # Check missing titles
    assert as_data.record(1, by_index=True).title == ""
    assert as_data.record(3, by_index=True).title == ""

    # Check missing abstracts
    assert as_data.record(0, by_index=True).abstract == ""
    assert as_data.record(2, by_index=True).abstract == ""

    # Check missing authors
    assert as_data.record(0, by_index=True).authors is None
    assert as_data.record(2, by_index=True).authors is None

    # Check missing keywords
    assert as_data.record(0, by_index=True).keywords is None
    assert as_data.record(2, by_index=True).keywords is None

    # Check missing notes
    assert as_data.record(0, by_index=True).notes is None
    assert as_data.record(2, by_index=True).notes is None

    # check missing doi
    assert as_data.record(0, by_index=True).doi is None
    assert as_data.record(2, by_index=True).doi is None


def test_nan_values_csv():
    fp = Path("tests", "demo_data", "missing_values.csv")
    as_data = ASReviewData.from_file(fp)

    # Check missing titles
    assert as_data.record(1, by_index=True).title == ""
    assert as_data.record(3, by_index=True).title == ""

    # Check missing abstracts
    assert as_data.record(0, by_index=True).abstract == ""
    assert as_data.record(2, by_index=True).abstract == ""

    # Check missing authors
    assert as_data.record(0, by_index=True).authors is None
    assert as_data.record(2, by_index=True).authors is None

    # Check missing keywords
    assert as_data.record(0, by_index=True).keywords is None
    assert as_data.record(2, by_index=True).keywords is None

    # Check missing doi
    assert as_data.record(0, by_index=True).doi is None
    assert as_data.record(2, by_index=True).doi is None


def test_write_data(tmpdir):
    fp_in = Path("tests", "demo_data", "generic_labels.csv")
    fp_out = Path(tmpdir, "generic_out.csv")
    asr_data = ASReviewData.from_file(fp_in)
    asr_data.to_file(fp_out, labels=[[0, 0], [2, 1], [3, 1]])

    tmp_csv_fp_out = Path(tmpdir, "tmp_generic_labels.csv")
    asr_data.to_file(tmp_csv_fp_out)
    asr_data_diff = ASReviewData.from_file(tmp_csv_fp_out)
    # Check if export file includes labels [1,0]
    assert list(asr_data.labels) == list(asr_data_diff.labels)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = []

from pathlib import Path

from asreview.state.errors import StateNotFoundError

V3STATE_VERSION = "1.0"


def is_zipped_project_file(fp):
    """Check if it is a zipped asreview project file."""
    if Path(fp).is_file():
        state_ext = Path(fp).suffix

        if state_ext in [".h5", ".hdf5", ".he5", ".json"]:
            raise ValueError(
                f"State file with extension {state_ext} is no longer "
                f"supported. Migrate to the new format or "
                "use an older version of ASReview. See LINK."
            )
        elif state_ext == ".asreview":
            return True
        else:
            raise ValueError(f"State file extension {state_ext} is not " f"recognized.")
    else:
        return False


def is_valid_project_folder(fp):
    """Check of the folder contains an asreview project."""
    if not Path(fp, "reviews").is_dir() or not Path(fp, "feature_matrices").is_dir():
        raise StateNotFoundError(
            f"There does not seem to be a valid project folder at {fp}. The "
            f"'reviews' or 'feature_matrices' folder is missing."
        )
    else:
        return



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# based on https://github.com/pypa/sampleproject - MIT License

# Always prefer setuptools over distutils
import platform
import re
import subprocess
from io import open
from os import path
from pathlib import Path

from setuptools import Command
from setuptools import find_packages
from setuptools import setup

import versioneer


def get_long_description():
    """Get project description based on README."""
    here = path.abspath(path.dirname(__file__))

    # Get the long description from the README file
    with open(path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = f.read()

    # remove emoji
    long_description = re.sub(r"\:[a-z_]+\:", "", long_description)

    return long_description


REQUIRES = [
    "numpy",
    "pandas>=2,<3",
    "scikit-learn",
    "rispy~=0.7.0",
    "setuptools",
    "flask>=2.3.0",
    "flask_cors",
    "flask-login==0.6.3",
    "flask-mail",
    "Werkzeug>=2.3.2",
    "openpyxl>=3.0.3",
    "xlsxwriter>=3",
    "jsonschema",
    "filelock",
    "Flask-SQLAlchemy>=3.0.2",
    "requests",
    "tqdm",
    "waitress",
    "datahugger>=0.2",
    "synergy_dataset",
    "sqlalchemy-utils",
    "tomli; python_version < '3.11'",
    "importlib_metadata; python_version < '3.10'",
]


DEPS = {
    "sbert": ["sentence_transformers"],
    "doc2vec": ["gensim"],
    "tensorflow": ["tensorflow~=2.0"],
    "dev": ["black", "check-manifest", "flake8", "flake8-isort", "isort"],
    "test": [
        "coverage",
        "pytest",
        "pytest-random-order",
        "pytest-selenium"
    ],
    "docs": [
        "ipython",
        "sphinx",
        "sphinx_rtd_theme",
        "sphinx-reredirects",
        "sphinxcontrib-youtube",
        "nbsphinx",
    ],
}
DEPS["all"] = DEPS["sbert"] + DEPS["doc2vec"]
DEPS["all"] += DEPS["tensorflow"]


class CompileAssets(Command):
    """
    Compile and build the frontend assets using npm and webpack.

    Registered as cmdclass in setup() so it can be called with
    ``python setup.py compile_assets``.
    """

    description = "Compile and build the frontend assets"
    user_options = []

    def initialize_options(self):
        """Set default values for options."""

    def finalize_options(self):
        """Set final values for options."""

    def run(self):
        """Run a command to compile and build assets."""

        path_webapp = Path(__file__).parent / "asreview" / "webapp"

        subprocess.check_call(
            ["npm", "install"],
            cwd=str(path_webapp),
            shell=(platform.system() == "Windows"),
        )
        subprocess.check_call(
            ["npm", "run-script", "build"],
            cwd=str(path_webapp),
            shell=(platform.system() == "Windows"),
        )


def get_cmdclass():
    cmdclass = versioneer.get_cmdclass()
    cmdclass["compile_assets"] = CompileAssets
    return cmdclass


setup(
    name="asreview",
    version=versioneer.get_version(),
    cmdclass=get_cmdclass(),
    description="ASReview LAB - A tool for AI-assisted systematic reviews",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/asreview/asreview",
    author="ASReview LAB developers",
    author_email="asreview@uu.nl",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Text Processing :: General",
        "Framework :: Flask",
    ],
    keywords=["systematic review", "machine-learning"],
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    package_data={
        "asreview": [
            "webapp/build/*",
            "webapp/build/static/*/*",
            "webapp/templates/emails/*",
        ]
    },
    python_requires="~=3.8",
    install_requires=REQUIRES,
    extras_require=DEPS,
    entry_points={
        "console_scripts": [
            "asreview=asreview.__main__:main",
        ],
        "asreview.entry_points": [
            "lab=asreview.webapp.entry_points.lab:lab_entry_point",
            "simulate=asreview.entry_points:SimulateEntryPoint",
            "algorithms=asreview.entry_points:AlgorithmsEntryPoint",
            "auth-tool=asreview.webapp.entry_points.auth_tool:AuthTool",
        ],
        "asreview.entry_points_internal": [
            "web_run_model=asreview.webapp.entry_points.run_model:run_model_entry_point",  # noqa
            "state-inspect=asreview.entry_points:StateInspectEntryPoint",
        ],
        "asreview.readers": [
            ".csv = asreview.io:CSVReader",
            ".tab = asreview.io:CSVReader",
            ".tsv = asreview.io:CSVReader",
            ".ris = asreview.io:RISReader",
            ".txt = asreview.io:RISReader",
            ".xlsx = asreview.io:ExcelReader",
        ],
        "asreview.writers": [
            ".csv = asreview.io:CSVWriter",
            ".tab = asreview.io:TSVWriter",
            ".tsv = asreview.io:TSVWriter",
            ".ris = asreview.io:RISWriter",
            ".txt = asreview.io:RISWriter",
            ".xlsx = asreview.io:ExcelWriter",
        ],
        "asreview.datasets": [
            "benchmark-nature = asreview.datasets:NaturePublicationDataGroup",
            "synergy = asreview.datasets:SynergyDataGroup",
            "benchmark = asreview.datasets:BenchmarkDataGroup",
        ],
        "asreview.models.classifiers": [
            "svm = asreview.models.classifiers:SVMClassifier",
            "nb = asreview.models.classifiers:NaiveBayesClassifier",
            "rf = asreview.models.classifiers:RandomForestClassifier",
            "nn-2-layer = asreview.models.classifiers:NN2LayerClassifier",
            "logistic = asreview.models.classifiers:LogisticClassifier",
            "lstm-base = asreview.models.classifiers:LSTMBaseClassifier",
            "lstm-pool = asreview.models.classifiers:LSTMPoolClassifier",
        ],
        "asreview.models.feature_extraction": [
            "doc2vec = asreview.models.feature_extraction:Doc2Vec",
            "embedding-idf = asreview.models.feature_extraction:EmbeddingIdf",
            "embedding-lstm = asreview.models.feature_extraction:EmbeddingLSTM",
            "sbert = asreview.models.feature_extraction:SBERT",
            "tfidf = asreview.models.feature_extraction:Tfidf",
        ],
        "asreview.models.balance": [
            "simple = asreview.models.balance:SimpleBalance",
            "double = asreview.models.balance:DoubleBalance",
            # "triple = asreview.models.balance:TripleBalance",  # Broken, only via API
            "undersample = asreview.models.balance:UndersampleBalance",
        ],
        "asreview.models.query": [
            "max = asreview.models.query.max:MaxQuery",
            "random = asreview.models.query.random:RandomQuery",
            "uncertainty = asreview.models.query.uncertainty:UncertaintyQuery",
            "cluster = asreview.models.query.cluster:ClusterQuery",
            "max_random = asreview.models.query.mixed:MaxRandomQuery",
            "max_uncertainty = asreview.models.query.mixed:MaxUncertaintyQuery",
        ],
    },
    project_urls={
        "Bug Reports": "https://github.com/asreview/asreview/issues",
        "Source": "https://github.com/asreview/asreview/",
    },
)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["BaseQueryStrategy", "ProbaQueryStrategy"]

from abc import abstractmethod

from asreview.models.base import BaseModel


class BaseQueryStrategy(BaseModel):
    """Abstract class for query strategies."""

    name = "base-query"

    @abstractmethod
    def query(
        self,
        X,
        classifier=None,
        n_instances=None,
        return_classifier_scores=False,
        **kwargs
    ):
        """Put records in ranked order.

        Arguments
        ---------
        X: numpy.ndarray
            Feature matrix where every row contains the features of a record.
        classifier: SKLearnModel
            Trained classifier to compute relevance scores.
        n_instances: int
            Number of records to query. If None returns all records in ranked order.
        return_classifier_score : bool
            Return the relevance scores produced by the classifier.

        Returns
        -------
        numpy.ndarray or (numpy.ndarray, np.ndarray)
            The QueryStrategy ranks the row numbers of the feature matrix. It returns
            an array of shape (n_instances,) containing the row indices in ranked
            order.
            If n_instances is None, returns all row numbers in ranked order. If
            n_instances is an integer, it only returns the top n_instances.
            If return_classifier_scores=True, also returns a second array with the same
            number of rows as the feature matrix, containing the relevance scores
            predicted by the classifier. If the classifier is not used, this will be
            None.
        """
        raise NotImplementedError


class ProbaQueryStrategy(BaseQueryStrategy):
    name = "proba"

    def query(
        self, X, classifier, n_instances=None, return_classifier_scores=False, **kwargs
    ):
        """Query method for strategies which use class probabilities."""
        if n_instances is None:
            n_instances = X.shape[0]

        predictions = classifier.predict_proba(X)

        query_idx = self._query(predictions, n_instances, X)

        if return_classifier_scores:
            return query_idx, predictions
        else:
            return query_idx

    @abstractmethod
    def _query(self, predictions, n_instances, X=None):
        raise NotImplementedError



import os

import pytest

from asreview import ASReviewData
from asreview.models.feature_extraction import get_feature_model
from asreview.models.feature_extraction import list_feature_extraction

ADVANCED_DEPS = {"tensorflow": False}

REQUIRES_EXTRA_DEPS = ["doc2vec", "embedding-idf", "sbert"]

try:
    import tensorflow  # noqa

    ADVANCED_DEPS["tensorflow"] = True
except ImportError:
    pass


@pytest.mark.parametrize(
    "feature_extraction",
    [
        "doc2vec",
        "embedding-idf",
        #  "sbert",
        "tfidf",
    ],
)
@pytest.mark.parametrize(
    "split_ta",
    [
        0,
        1,
    ],
)
def test_features(feature_extraction, split_ta):
    if feature_extraction in REQUIRES_EXTRA_DEPS and not ADVANCED_DEPS["tensorflow"]:
        pytest.skip()

    embedding_fp = os.path.join("tests", "demo_data", "generic.vec")
    data_fp = os.path.join("tests", "demo_data", "generic.csv")

    as_data = ASReviewData.from_file(data_fp)
    texts = as_data.texts
    if feature_extraction.startswith("embedding-"):
        model = get_feature_model(
            feature_extraction, split_ta=split_ta, embedding_fp=embedding_fp
        )
    else:
        model = get_feature_model(feature_extraction, split_ta=split_ta)
    X = model.fit_transform(texts, titles=as_data.title, abstracts=as_data.abstract)

    assert X.shape[0] == len(as_data.title)
    assert X.shape[1] > 0
    assert isinstance(model.param, dict)
    assert model.name == feature_extraction


def test_feature_general():
    assert len(list_feature_extraction()) >= 5



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["LABEntryPoint"]

import os

from asreview.webapp.entry_points.lab import lab_entry_point


def _deprecated_dev_mode():
    if os.environ.get("FLASK_DEBUG", "") == "1":
        print(
            "\n\n\n!IMPORTANT!\n\n"
            "asreview lab development mode is deprecated, see:\n"
            "https://github.com/J535D165/asreview/blob/master/DEVELOPMENT.md"
            "\n\n\n"
        )
        exit(1)


class LABEntryPoint(object):
    """Entry point to start the ASReview LAB webapp."""

    def execute(self, argv):
        # check deprecated dev mode
        _deprecated_dev_mode()

        lab_entry_point(argv)
